#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import os
import sys
import unittest
import yaml
import re

from pkg_resources import parse_version


class TestYaml(unittest.TestCase):
    filename = ''

    @classmethod
    def setUpClass(cls):
        with open(cls.filename) as f:
            cls.yaml = yaml.safe_load(f)

    def dict_key_crosscheck(self, d, keys):
        '''Check that each key in d is in keys, and that each key is in d'''
        self.assertEqual(sorted(d.keys()), sorted(keys))

    def libinput_events(self, filter=None):
        '''Returns all libinput events in the recording, regardless of the
        device'''
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    libinput = e['libinput']
                except KeyError:
                    continue

                for ev in libinput:
                    if (filter is None or ev['type'] == filter or
                            isinstance(filter, list) and ev['type'] in filter):
                        yield ev

    def test_sections_exist(self):
        sections = ['version', 'ndevices', 'libinput', 'system', 'devices']
        for section in sections:
            self.assertIn(section, self.yaml)

    def test_version(self):
        version = self.yaml['version']
        self.assertTrue(isinstance(version, int))
        self.assertEqual(version, 1)

    def test_ndevices(self):
        ndevices = self.yaml['ndevices']
        self.assertTrue(isinstance(ndevices, int))
        self.assertGreaterEqual(ndevices, 1)
        self.assertEqual(ndevices, len(self.yaml['devices']))

    def test_libinput(self):
        libinput = self.yaml['libinput']
        version = libinput['version']
        self.assertTrue(isinstance(version, str))
        self.assertGreaterEqual(parse_version(version),
                                parse_version('1.10.0'))
        git = libinput['git']
        self.assertTrue(isinstance(git, str))
        self.assertNotEqual(git, 'unknown')

    def test_system(self):
        system = self.yaml['system']
        kernel = system['kernel']
        self.assertTrue(isinstance(kernel, str))
        self.assertEqual(kernel, os.uname().release)

        dmi = system['dmi']
        self.assertTrue(isinstance(dmi, str))
        with open('/sys/class/dmi/id/modalias') as f:
            sys_dmi = f.read()[:-1]  # trailing newline
            self.assertEqual(dmi, sys_dmi)

    def test_devices_sections_exist(self):
        devices = self.yaml['devices']
        for d in devices:
            self.assertIn('node', d)
            self.assertIn('evdev', d)
            self.assertIn('udev', d)

    def test_evdev_sections_exist(self):
        sections = ['name', 'id', 'codes', 'properties']
        devices = self.yaml['devices']
        for d in devices:
            evdev = d['evdev']
            for s in sections:
                self.assertIn(s, evdev)

    def test_evdev_name(self):
        devices = self.yaml['devices']
        for d in devices:
            evdev = d['evdev']
            name = evdev['name']
            self.assertTrue(isinstance(name, str))
            self.assertGreaterEqual(len(name), 5)

    def test_evdev_id(self):
        devices = self.yaml['devices']
        for d in devices:
            evdev = d['evdev']
            id = evdev['id']
            self.assertTrue(isinstance(id, list))
            self.assertEqual(len(id), 4)
            self.assertGreater(id[0], 0)
            self.assertGreater(id[1], 0)

    def test_evdev_properties(self):
        devices = self.yaml['devices']
        for d in devices:
            evdev = d['evdev']
            properties = evdev['properties']
            self.assertTrue(isinstance(properties, list))

    def test_hid(self):
        devices = self.yaml['devices']
        for d in devices:
            hid = d['hid']
            self.assertTrue(isinstance(hid, list))
            for byte in hid:
                self.assertGreaterEqual(byte, 0)
                self.assertLessEqual(byte, 255)

    def test_udev_sections_exist(self):
        sections = ['properties']
        devices = self.yaml['devices']
        for d in devices:
            udev = d['udev']
            for s in sections:
                self.assertIn(s, udev)

    def test_udev_properties(self):
        devices = self.yaml['devices']
        for d in devices:
            udev = d['udev']
            properties = udev['properties']
            self.assertTrue(isinstance(properties, list))
            self.assertGreater(len(properties), 0)

            self.assertIn('ID_INPUT=1', properties)
            for p in properties:
                self.assertTrue(re.match('[A-Z0-9_]+=.+', p))

    def test_udev_id_inputs(self):
        devices = self.yaml['devices']
        for d in devices:
            udev = d['udev']
            properties = udev['properties']
            id_inputs = [p for p in properties if p.startswith('ID_INPUT')]
            # We expect ID_INPUT and ID_INPUT_something, but might get more
            # than one of the latter
            self.assertGreaterEqual(len(id_inputs), 2)

    def test_events_have_section(self):
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                self.assertTrue('evdev' in e or 'libinput' in e)

    def test_events_evdev(self):
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e['evdev']
                except KeyError:
                    continue

                for ev in evdev:
                    self.assertEqual(len(ev), 5)

                # Last event in each frame is SYN_REPORT
                ev_syn = evdev[-1]
                self.assertEqual(ev_syn[2], 0)
                self.assertEqual(ev_syn[3], 0)
                # SYN_REPORT value is 1 in case of some key repeats
                self.assertLessEqual(ev_syn[4], 1)

    def test_events_evdev_syn_report(self):
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e['evdev']
                except KeyError:
                    continue
                for ev in evdev[:-1]:
                    self.assertFalse(ev[2] == 0 and ev[3] == 0)

    def test_events_libinput(self):
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    libinput = e['libinput']
                except KeyError:
                    continue

                self.assertTrue(isinstance(libinput, list))
                for ev in libinput:
                    self.assertTrue(isinstance(ev, dict))

    def test_events_libinput_type(self):
        types = ['POINTER_MOTION', 'POINTER_MOTION_ABSOLUTE', 'POINTER_AXIS',
                 'POINTER_BUTTON', 'DEVICE_ADDED', 'KEYBOARD_KEY',
                 'TOUCH_DOWN', 'TOUCH_MOTION', 'TOUCH_UP', 'TOUCH_FRAME',
                 'GESTURE_SWIPE_BEGIN', 'GESTURE_SWIPE_UPDATE',
                 'GESTURE_SWIPE_END', 'GESTURE_PINCH_BEGIN',
                 'GESTURE_PINCH_UPDATE', 'GESTURE_PINCH_END',
                 'TABLET_TOOL_AXIS', 'TABLET_TOOL_PROXIMITY',
                 'TABLET_TOOL_BUTTON', 'TABLET_TOOL_TIP',
                 'TABLET_PAD_STRIP', 'TABLET_PAD_RING',
                 'TABLET_PAD_BUTTON', 'SWITCH_TOGGLE',
                 ]
        for e in self.libinput_events():
            self.assertIn('type', e)
            self.assertIn(e['type'], types)

    def test_events_libinput_time(self):
        # DEVICE_ADDED has no time
        # first event may have 0.0 time if the first frame generates a
        # libinput event.
        try:
            for e in list(self.libinput_events())[2:]:
                self.assertIn('time', e)
                self.assertGreater(e['time'], 0.0)
                self.assertLess(e['time'], 60.0)
        except IndexError:
            pass

    def test_events_libinput_device_added(self):
        keys = ['type', 'seat', 'logical_seat']
        for e in self.libinput_events('DEVICE_ADDED'):
            self.dict_key_crosscheck(e, keys)
            self.assertEqual(e['seat'], 'seat0')
            self.assertEqual(e['logical_seat'], 'default')

    def test_events_libinput_pointer_motion(self):
        keys = ['type', 'time', 'delta', 'unaccel']
        for e in self.libinput_events('POINTER_MOTION'):
            self.dict_key_crosscheck(e, keys)
            delta = e['delta']
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e['unaccel']
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_pointer_button(self):
        keys = ['type', 'time', 'button', 'state', 'seat_count']
        for e in self.libinput_events('POINTER_BUTTON'):
            self.dict_key_crosscheck(e, keys)
            button = e['button']
            self.assertGreater(button, 0x100)  # BTN_0
            self.assertLess(button, 0x160)  # KEY_OK
            state = e['state']
            self.assertIn(state, ['pressed', 'released'])
            scount = e['seat_count']
            self.assertGreaterEqual(scount, 0)

    def test_events_libinput_pointer_absolute(self):
        keys = ['type', 'time', 'point', 'transformed']
        for e in self.libinput_events('POINTER_MOTION_ABSOLUTE'):
            self.dict_key_crosscheck(e, keys)
            point = e['point']
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e['transformed']
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch(self):
        keys = ['type', 'time', 'slot', 'seat_slot']
        for e in self.libinput_events():
            if (not e['type'].startswith('TOUCH_') or
                    e['type'] == 'TOUCH_FRAME'):
                continue

            for k in keys:
                self.assertIn(k, e.keys())
            slot = e['slot']
            seat_slot = e['seat_slot']

            self.assertGreaterEqual(slot, 0)
            self.assertGreaterEqual(seat_slot, 0)

    def test_events_libinput_touch_down(self):
        keys = ['type', 'time', 'slot', 'seat_slot', 'point', 'transformed']
        for e in self.libinput_events('TOUCH_DOWN'):
            self.dict_key_crosscheck(e, keys)
            point = e['point']
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e['transformed']
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch_motion(self):
        keys = ['type', 'time', 'slot', 'seat_slot', 'point', 'transformed']
        for e in self.libinput_events('TOUCH_MOTION'):
            self.dict_key_crosscheck(e, keys)
            point = e['point']
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e['transformed']
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch_frame(self):
        devices = self.yaml['devices']
        for d in devices:
            events = d['events']
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e['libinput']
                except KeyError:
                    continue

                need_frame = False
                for ev in evdev:
                    t = ev['type']
                    if not t.startswith('TOUCH_'):
                        self.assertFalse(need_frame)
                        continue

                    if t == 'TOUCH_FRAME':
                        self.assertTrue(need_frame)
                        need_frame = False
                    else:
                        need_frame = True

                self.assertFalse(need_frame)

    def test_events_libinput_gesture_pinch(self):
        keys = ['type', 'time', 'nfingers', 'delta',
                'unaccel', 'angle_delta', 'scale']
        for e in self.libinput_events(['GESTURE_PINCH_BEGIN',
                                       'GESTURE_PINCH_UPDATE',
                                       'GESTURE_PINCH_END']):
            self.dict_key_crosscheck(e, keys)
            delta = e['delta']
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e['unaccel']
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

            adelta = e['angle_delta']
            self.assertTrue(isinstance(adelta, list))
            self.assertEqual(len(adelta), 2)
            for d in adelta:
                self.assertTrue(isinstance(d, float))

            scale = e['scale']
            self.assertTrue(isinstance(scale, list))
            self.assertEqual(len(scale), 2)
            for d in scale:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_gesture_swipe(self):
        keys = ['type', 'time', 'nfingers', 'delta',
                'unaccel']
        for e in self.libinput_events(['GESTURE_SWIPE_BEGIN',
                                       'GESTURE_SWIPE_UPDATE',
                                       'GESTURE_SWIPE_END']):
            self.dict_key_crosscheck(e, keys)
            delta = e['delta']
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e['unaccel']
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_tablet_pad_button(self):
        keys = ['type', 'time', 'button', 'state', 'mode', 'is-toggle']

        for e in self.libinput_events('TABLET_PAD_BUTTON'):
            self.dict_key_crosscheck(e, keys)

            b = e['button']
            self.assertTrue(isinstance(b, int))
            self.assertGreaterEqual(b, 0)
            self.assertLessEqual(b, 16)

            state = e['state']
            self.assertIn(state, ['pressed', 'released'])

            m = e['mode']
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            t = e['is-toggle']
            self.assertTrue(isinstance(t, bool))

    def test_events_libinput_tablet_pad_ring(self):
        keys = ['type', 'time', 'number', 'position', 'source', 'mode']

        for e in self.libinput_events('TABLET_PAD_RING'):
            self.dict_key_crosscheck(e, keys)

            n = e['number']
            self.assertTrue(isinstance(n, int))
            self.assertGreaterEqual(n, 0)
            self.assertLessEqual(n, 4)

            p = e['position']
            self.assertTrue(isinstance(p, float))
            if p != -1.0:  # special 'end' case
                self.assertGreaterEqual(p, 0.0)
                self.assertLess(p, 360.0)

            m = e['mode']
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            s = e['source']
            self.assertIn(s, ['finger', 'unknown'])

    def test_events_libinput_tablet_pad_strip(self):
        keys = ['type', 'time', 'number', 'position', 'source', 'mode']

        for e in self.libinput_events('TABLET_PAD_STRIP'):
            self.dict_key_crosscheck(e, keys)

            n = e['number']
            self.assertTrue(isinstance(n, int))
            self.assertGreaterEqual(n, 0)
            self.assertLessEqual(n, 4)

            p = e['position']
            self.assertTrue(isinstance(p, float))
            if p != -1.0:  # special 'end' case
                self.assertGreaterEqual(p, 0.0)
                self.assertLessEqual(p, 1.0)

            m = e['mode']
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            s = e['source']
            self.assertIn(s, ['finger', 'unknown'])

    def test_events_libinput_tablet_tool_proximity(self):
        keys = ['type', 'time', 'proximity', 'tool-type', 'serial', 'axes']

        for e in self.libinput_events('TABLET_TOOL_PROXIMITY'):
            for k in keys:
                self.assertIn(k, e)

            p = e['proximity']
            self.assertIn(p, ['in', 'out'])

            p = e['tool-type']
            self.assertIn(p, ['pen', 'eraser', 'brush', 'airbrush', 'mouse',
                              'lens', 'unknown'])

            s = e['serial']
            self.assertTrue(isinstance(s, int))
            self.assertGreaterEqual(s, 0)

            a = e['axes']
            for ax in e['axes']:
                self.assertIn(a, 'pdtrsw')

    def test_events_libinput_tablet_tool(self):
        keys = ['type', 'time', 'tip']

        for e in self.libinput_events(['TABLET_TOOL_AXIS',
                                       'TABLET_TOOL_TIP']):
            for k in keys:
                self.assertIn(k, e)

            t = e['tip']
            self.assertIn(t, ['down', 'up'])

    def test_events_libinput_tablet_tool_button(self):
        keys = ['type', 'time', 'button', 'state']

        for e in self.libinput_events('TABLET_TOOL_BUTTON'):
            self.dict_key_crosscheck(e, keys)

            b = e['button']
            # STYLUS, STYLUS2, STYLUS3
            self.assertIn(b, [0x14b, 0x14c, 0x139])

            s = e['state']
            self.assertIn(s, ['pressed', 'released'])

    def test_events_libinput_tablet_tool_axes(self):
        for e in self.libinput_events(['TABLET_TOOL_PROXIMITY',
                                       'TABLET_TOOL_AXIS',
                                       'TABLET_TOOL_TIP']):

            point = e['point']
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)

            try:
                tilt = e['tilt']
                self.assertTrue(isinstance(tilt, list))
                self.assertEqual(len(tilt), 2)
                for t in tilt:
                    self.assertTrue(isinstance(t, float))
            except KeyError:
                pass

            try:
                d = e['distance']
                self.assertTrue(isinstance(d, float))
                self.assertGreaterEqual(d, 0.0)
                self.assertNotIn('pressure', e)
            except KeyError:
                pass

            try:
                p = e['pressure']
                self.assertTrue(isinstance(p, float))
                self.assertGreaterEqual(p, 0.0)
                self.assertNotIn('distance', e)
            except KeyError:
                pass

            try:
                r = e['rotation']
                self.assertTrue(isinstance(r, float))
                self.assertGreaterEqual(r, 0.0)
            except KeyError:
                pass

            try:
                s = e['slider']
                self.assertTrue(isinstance(s, float))
                self.assertGreaterEqual(s, 0.0)
            except KeyError:
                pass

            try:
                w = e['wheel']
                self.assertTrue(isinstance(w, float))
                self.assertGreaterEqual(w, 0.0)
                self.assertIn('wheel-discrete', e)
                wd = e['wheel-discrete']
                self.assertTrue(isinstance(wd, 1))
                self.assertGreaterEqual(wd, 0.0)

                def sign(x):
                    (1, -1)[x < 0]
                self.assertTrue(sign(w), sign(wd))
            except KeyError:
                pass

    def test_events_libinput_switch(self):
        keys = ['type', 'time', 'switch', 'state']

        for e in self.libinput_events('SWITCH_TOGGLE'):
            self.dict_key_crosscheck(e, keys)

            s = e['switch']
            self.assertTrue(isinstance(s, int))
            self.assertIn(s, [0x00, 0x01])

            # yaml converts on/off to true/false
            state = e['state']
            self.assertTrue(isinstance(state, bool))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Verify a YAML recording')
    parser.add_argument('recording', metavar='recorded-file.yaml',
                        type=str, help='Path to device recording')
    parser.add_argument('--verbose', action='store_true')
    args, remainder = parser.parse_known_args()
    TestYaml.filename = args.recording
    verbosity = 1
    if args.verbose:
        verbosity = 3

    argv = [sys.argv[0], *remainder]
    unittest.main(argv=argv, verbosity=verbosity)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#

import sys
import subprocess
import argparse
try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print('Error: {}'.format(str(e)), file=sys.stderr)
    print('One or more python modules are missing. Please install those '
          'modules and re-run this tool.')
    sys.exit(1)


class Range(object):
    """Class to keep a min/max of a value around"""
    def __init__(self):
        self.min = float('inf')
        self.max = float('-inf')

    def update(self, value):
        self.min = min(self.min, value)
        self.max = max(self.max, value)


class Touch(object):
    """A single data point of a sequence (i.e. one event frame)"""

    def __init__(self, major=None, minor=None, orientation=None):
        self._major = major
        self._minor = minor
        self._orientation = orientation
        self.dirty = False

    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, major):
        self._major = major
        self.dirty = True

    @property
    def minor(self):
        return self._minor

    @minor.setter
    def minor(self, minor):
        self._minor = minor
        self.dirty = True

    @property
    def orientation(self):
        return self._orientation

    @orientation.setter
    def orientation(self, orientation):
        self._orientation = orientation
        self.dirty = True

    def __str__(self):
        s = "Touch: major {:3d}".format(self.major)
        if self.minor is not None:
            s += ", minor {:3d}".format(self.minor)
        if self.orientation is not None:
            s += ", orientation {:+3d}".format(self.orientation)
        return s


class TouchSequence(object):
    """A touch sequence from beginning to end"""

    def __init__(self, device, tracking_id):
        self.device = device
        self.tracking_id = tracking_id
        self.points = []

        self.is_active = True

        self.is_down = False
        self.was_down = False
        self.is_palm = False
        self.was_palm = False
        self.is_thumb = False
        self.was_thumb = False

        self.major_range = Range()
        self.minor_range = Range()

    def append(self, touch):
        """Add a Touch to the sequence"""
        self.points.append(touch)
        self.major_range.update(touch.major)
        self.minor_range.update(touch.minor)

        if touch.major < self.device.up or touch.minor < self.device.up:
            self.is_down = False
        elif touch.major > self.device.down or touch.minor > self.device.down:
            self.is_down = True
            self.was_down = True

        self.is_palm = touch.major > self.device.palm
        if self.is_palm:
            self.was_palm = True

        self.is_thumb = self.device.thumb != 0 and touch.major > self.device.thumb
        if self.is_thumb:
            self.was_thumb = True

    def finalize(self):
        """Mark the TouchSequence as complete (finger is up)"""
        self.is_active = False

    def __str__(self):
        return self._str_state() if self.is_active else self._str_summary()

    def _str_summary(self):
        if not self.points:
            return "{:78s}".format("Sequence: no major/minor values recorded")

        s = "Sequence: major: [{:3d}..{:3d}] ".format(
            self.major_range.min, self.major_range.max
        )
        if self.device.has_minor:
            s += "minor: [{:3d}..{:3d}] ".format(self.minor_range.min, self.minor_range.max)
        if self.was_down:
            s += " down"
        if self.was_palm:
            s += " palm"
        if self.was_thumb:
            s += " thumb"

        return s

    def _str_state(self):
        touch = self.points[-1]
        s = "{}, tags: {} {} {}".format(touch,
                                        "down" if self.is_down else "    ",
                                        "palm" if self.is_palm else "    ",
                                        "thumb" if self.is_thumb else "     ")
        return s


class InvalidDeviceError(Exception):
    pass


class Device(libevdev.Device):
    def __init__(self, path):
        if path is None:
            self.path = self.find_touch_device()
        else:
            self.path = path

        fd = open(self.path, 'rb')
        super().__init__(fd)

        print("Using {}: {}\n".format(self.name, self.path))

        if not self.has(libevdev.EV_ABS.ABS_MT_TOUCH_MAJOR):
            raise InvalidDeviceError("Device does not have ABS_MT_TOUCH_MAJOR")

        self.has_minor = self.has(libevdev.EV_ABS.ABS_MT_TOUCH_MINOR)
        self.has_orientation = self.has(libevdev.EV_ABS.ABS_MT_ORIENTATION)

        self.up = 0
        self.down = 0
        self.palm = 0
        self.thumb = 0

        self._init_thresholds_from_quirks()
        self.sequences = []
        self.touch = Touch(0, 0)

    def find_touch_device(self):
        context = pyudev.Context()
        for device in context.list_devices(subsystem='input'):
            if not device.get('ID_INPUT_TOUCHPAD', 0) and \
               not device.get('ID_INPUT_TOUCHSCREEN', 0):
                continue

            if not device.device_node or \
                    not device.device_node.startswith('/dev/input/event'):
                continue

            return device.device_node

        print("Unable to find a touch device.", file=sys.stderr)
        sys.exit(1)

    def _init_thresholds_from_quirks(self):
        command = ['libinput', 'quirks', 'list', self.path]
        cmd = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            print("Error querying quirks: {}".format(cmd.stderr.decode('utf-8')), file=sys.stderr)
            return

        stdout = cmd.stdout.decode('utf-8')
        quirks = [q.split('=') for q in stdout.split('\n')]

        for q in quirks:
            if q[0] == 'AttrPalmSizeThreshold':
                self.palm = int(q[1])
            elif q[0] == 'AttrTouchSizeRange':
                self.down, self.up = colon_tuple(q[1])
            elif q[0] == 'AttrThumbSizeThreshold':
                self.thumb = int(q[1])

    def start_new_sequence(self, tracking_id):
        self.sequences.append(TouchSequence(self, tracking_id))

    def current_sequence(self):
        return self.sequences[-1]

    def handle_key(self, event):
        tapcodes = [libevdev.EV_KEY.BTN_TOOL_DOUBLETAP,
                    libevdev.EV_KEY.BTN_TOOL_TRIPLETAP,
                    libevdev.EV_KEY.BTN_TOOL_QUADTAP,
                    libevdev.EV_KEY.BTN_TOOL_QUINTTAP]
        if event.code in tapcodes and event.value > 0:
            print("\rThis tool cannot handle multiple fingers, "
                  "output will be invalid", file=sys.stderr)

    def handle_abs(self, event):
        if event.matches(libevdev.EV_ABS.ABS_MT_TRACKING_ID):
            if event.value > -1:
                self.start_new_sequence(event.value)
            else:
                try:
                    s = self.current_sequence()
                    s.finalize()
                    print("\r{}".format(s))
                except IndexError:
                    # If the finger was down during start
                    pass
        elif event.matches(libevdev.EV_ABS.ABS_MT_TOUCH_MAJOR):
            self.touch.major = event.value
        elif event.matches(libevdev.EV_ABS.ABS_MT_TOUCH_MINOR):
            self.touch.minor = event.value
        elif event.matches(libevdev.EV_ABS.ABS_MT_ORIENTATION):
            self.touch.orientation = event.value

    def handle_syn(self, event):
        if self.touch.dirty:
            try:
                self.current_sequence().append(self.touch)
                print("\r{}".format(self.current_sequence()), end="")
                self.touch = Touch(major=self.touch.major,
                                   minor=self.touch.minor,
                                   orientation=self.touch.orientation)
            except IndexError:
                pass

    def handle_event(self, event):
        if event.matches(libevdev.EV_ABS):
            self.handle_abs(event)
        elif event.matches(libevdev.EV_KEY):
            self.handle_key(event)
        elif event.matches(libevdev.EV_SYN):
            self.handle_syn(event)

    def read_events(self):
        print("Ready for recording data.")
        print("Touch sizes used: {}:{}".format(self.down, self.up))
        print("Palm size used: {}".format(self.palm))
        print("Thumb size used: {}".format(self.thumb))
        print("Place a single finger on the device to measure touch size.\n"
              "Ctrl+C to exit\n")

        while True:
            for event in self.events():
                self.handle_event(event)


def colon_tuple(string):
    try:
        ts = string.split(':')
        t = tuple([int(x) for x in ts])
        if len(t) == 2 and t[0] >= t[1]:
            return t
    except: # noqa
        pass

    msg = "{} is not in format N:M (N >= M)".format(string)
    raise argparse.ArgumentTypeError(msg)


def main(args):
    parser = argparse.ArgumentParser(description="Measure touch size and orientation")
    parser.add_argument('path', metavar='/dev/input/event0',
                        nargs='?', type=str, help='Path to device (optional)')
    parser.add_argument('--touch-thresholds', metavar='down:up',
                        type=colon_tuple,
                        help='Thresholds when a touch is logically down or up')
    parser.add_argument('--palm-threshold', metavar='t',
                        type=int, help='Threshold when a touch is a palm')
    args = parser.parse_args()

    try:
        device = Device(args.path)

        if args.touch_thresholds is not None:
            device.down, device.up = args.touch_thresholds

        if args.palm_threshold is not None:
            device.palm = args.palm_threshold

        device.read_events()
    except KeyboardInterrupt:
        pass
    except (PermissionError, OSError):
        print("Error: failed to open device")
    except InvalidDeviceError as e:
        print("This device does not have the capabilities for size-based touch detection.")
        print("Details: {}".format(e))


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#

import sys
import subprocess
import argparse
try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print('Error: {}'.format(str(e)), file=sys.stderr)
    print('One or more python modules are missing. Please install those '
          'modules and re-run this tool.')
    sys.exit(1)


class Range(object):
    """Class to keep a min/max of a value around"""
    def __init__(self):
        self.min = float('inf')
        self.max = float('-inf')

    def update(self, value):
        self.min = min(self.min, value)
        self.max = max(self.max, value)


class Touch(object):
    """A single data point of a sequence (i.e. one event frame)"""

    def __init__(self, pressure=None):
        self.pressure = pressure


class TouchSequence(object):
    """A touch sequence from beginning to end"""

    def __init__(self, device, tracking_id):
        self.device = device
        self.tracking_id = tracking_id
        self.points = []

        self.is_active = True

        self.is_down = False
        self.was_down = False
        self.is_palm = False
        self.was_palm = False
        self.is_thumb = False
        self.was_thumb = False

        self.prange = Range()

    def append(self, touch):
        """Add a Touch to the sequence"""
        self.points.append(touch)
        self.prange.update(touch.pressure)

        if touch.pressure < self.device.up:
            self.is_down = False
        elif touch.pressure > self.device.down:
            self.is_down = True
            self.was_down = True

        self.is_palm = touch.pressure > self.device.palm
        if self.is_palm:
            self.was_palm = True

        self.is_thumb = touch.pressure > self.device.thumb
        if self.is_thumb:
            self.was_thumb = True

    def finalize(self):
        """Mark the TouchSequence as complete (finger is up)"""
        self.is_active = False

    def avg(self):
        """Average pressure value of this sequence"""
        return int(sum([p.pressure for p in self.points]) / len(self.points))

    def median(self):
        """Median pressure value of this sequence"""
        ps = sorted([p.pressure for p in self.points])
        idx = int(len(self.points) / 2)
        return ps[idx]

    def __str__(self):
        return self._str_state() if self.is_active else self._str_summary()

    def _str_summary(self):
        if not self.points:
            return "{:78s}".format("Sequence: no pressure values recorded")

        s = "Sequence {} pressure: "\
            "min: {:3d} max: {:3d} avg: {:3d} median: {:3d} tags:" \
            .format(
                self.tracking_id,
                self.prange.min,
                self.prange.max,
                self.avg(),
                self.median()
            )
        if self.was_down:
            s += " down"
        if self.was_palm:
            s += " palm"
        if self.was_thumb:
            s += " thumb"

        return s

    def _str_state(self):
        s = "Touchpad pressure: {:3d} min: {:3d} max: {:3d} tags: {} {} {}" \
            .format(
                self.points[-1].pressure,
                self.prange.min,
                self.prange.max,
                "down" if self.is_down else "    ",
                "palm" if self.is_palm else "    ",
                "thumb" if self.is_thumb else "     "
            )
        return s


class InvalidDeviceError(Exception):
    pass


class Device(libevdev.Device):
    def __init__(self, path):
        if path is None:
            self.path = self.find_touchpad_device()
        else:
            self.path = path

        fd = open(self.path, 'rb')
        super().__init__(fd)

        print("Using {}: {}\n".format(self.name, self.path))

        self.has_mt_pressure = True
        absinfo = self.absinfo[libevdev.EV_ABS.ABS_MT_PRESSURE]
        if absinfo is None:
            absinfo = self.absinfo[libevdev.EV_ABS.ABS_PRESSURE]
            self.has_mt_pressure = False
            if absinfo is None:
                raise InvalidDeviceError("Device does not have ABS_PRESSURE or ABS_MT_PRESSURE")

        prange = absinfo.maximum - absinfo.minimum

        # libinput defaults
        self.down = int(absinfo.minimum + 0.12 * prange)
        self.up = int(absinfo.minimum + 0.10 * prange)
        self.palm = 130  # the libinput default
        self.thumb = absinfo.maximum

        self._init_thresholds_from_quirks()
        self.sequences = []

    def find_touchpad_device(self):
        context = pyudev.Context()
        for device in context.list_devices(subsystem='input'):
            if not device.get('ID_INPUT_TOUCHPAD', 0):
                continue

            if not device.device_node or \
               not device.device_node.startswith('/dev/input/event'):
                continue

            return device.device_node
        print("Unable to find a touchpad device.", file=sys.stderr)
        sys.exit(1)

    def _init_thresholds_from_quirks(self):
        command = ['libinput', 'quirks', 'list', self.path]
        cmd = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            print("Error querying quirks: {}".format(cmd.stderr.decode('utf-8')), file=sys.stderr)
            return

        stdout = cmd.stdout.decode('utf-8')
        quirks = [q.split('=') for q in stdout.split('\n')]

        for q in quirks:
            if q[0] == 'AttrPalmPressureThreshold':
                self.palm = int(q[1])
            elif q[0] == 'AttrPressureRange':
                self.down, self.up = colon_tuple(q[1])
            elif q[0] == 'AttrThumbPressureThreshold':
                self.thumb = int(q[1])

    def start_new_sequence(self, tracking_id):
        self.sequences.append(TouchSequence(self, tracking_id))

    def current_sequence(self):
        return self.sequences[-1]


def handle_key(device, event):
    tapcodes = [
        libevdev.EV_KEY.BTN_TOOL_DOUBLETAP,
        libevdev.EV_KEY.BTN_TOOL_TRIPLETAP,
        libevdev.EV_KEY.BTN_TOOL_QUADTAP,
        libevdev.EV_KEY.BTN_TOOL_QUINTTAP
    ]
    if event.code in tapcodes and event.value > 0:
        print("\rThis tool cannot handle multiple fingers, "
              "output will be invalid", file=sys.stderr)


def handle_abs(device, event):
    if event.matches(libevdev.EV_ABS.ABS_MT_TRACKING_ID):
        if event.value > -1:
            device.start_new_sequence(event.value)
        else:
            try:
                s = device.current_sequence()
                s.finalize()
                print("\r{}".format(s))
            except IndexError:
                # If the finger was down at startup
                pass
    elif (event.matches(libevdev.EV_ABS.ABS_MT_PRESSURE) or
          (event.matches(libevdev.EV_ABS.ABS_PRESSURE) and not device.has_mt_pressure)):
        try:
            s = device.current_sequence()
            s.append(Touch(pressure=event.value))
            print("\r{}".format(s), end="")
        except IndexError:
            # If the finger was down at startup
            pass


def handle_event(device, event):
    if event.matches(libevdev.EV_ABS):
        handle_abs(device, event)
    elif event.matches(libevdev.EV_KEY):
        handle_key(device, event)


def loop(device):
    print("Ready for recording data.")
    print("Pressure range used: {}:{}".format(device.down, device.up))
    print("Palm pressure range used: {}".format(device.palm))
    print("Thumb pressure range used: {}".format(device.thumb))
    print("Place a single finger on the touchpad to measure pressure values.\n"
          "Ctrl+C to exit\n")

    while True:
        for event in device.events():
            handle_event(device, event)


def colon_tuple(string):
    try:
        ts = string.split(':')
        t = tuple([int(x) for x in ts])
        if len(t) == 2 and t[0] >= t[1]:
            return t
    except: # noqa
        pass

    msg = "{} is not in format N:M (N >= M)".format(string)
    raise argparse.ArgumentTypeError(msg)


def main(args):
    parser = argparse.ArgumentParser(
        description="Measure touchpad pressure values"
    )
    parser.add_argument(
        'path', metavar='/dev/input/event0', nargs='?', type=str,
        help='Path to device (optional)'
    )
    parser.add_argument(
        '--touch-thresholds', metavar='down:up', type=colon_tuple,
        help='Thresholds when a touch is logically down or up'
    )
    parser.add_argument(
        '--palm-threshold', metavar='t', type=int,
        help='Threshold when a touch is a palm'
    )
    args = parser.parse_args()

    try:
        device = Device(args.path)

        if args.touch_thresholds is not None:
            device.down, device.up = args.touch_thresholds

        if args.palm_threshold is not None:
            device.palm = args.palm_threshold

        loop(device)
    except KeyboardInterrupt:
        pass
    except (PermissionError, OSError):
        print("Error: failed to open device")
    except InvalidDeviceError as e:
        print("This device does not have the capabilities for pressure-based touch detection.")
        print("Details: {}".format(e))


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import os
import unittest
import resource
import sys
import subprocess
import tempfile
from pathlib import Path


def _disable_coredump():
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))


def run_command(args):
    with subprocess.Popen(args, preexec_fn=_disable_coredump,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE) as p:
        try:
            p.wait(0.7)
        except subprocess.TimeoutExpired:
            p.send_signal(3)  # SIGQUIT
        stdout, stderr = p.communicate(timeout=5)
        if p.returncode == -3:
            p.returncode = 0
        return p.returncode, stdout.decode('UTF-8'), stderr.decode('UTF-8')


class TestLibinputTool(unittest.TestCase):
    libinput_tool = 'libinput'
    subtool = None

    def run_command(self, args):
        args = [self.libinput_tool] + args
        if self.subtool is not None:
            args.insert(1, self.subtool)

        return run_command(args)

    def run_command_success(self, args):
        rc, stdout, stderr = self.run_command(args)
        # if we're running as user, we might fail the command but we should
        # never get rc 2 (invalid usage)
        self.assertIn(rc, [0, 1], msg=(stdout, stderr))

    def run_command_unrecognized_option(self, args):
        rc, stdout, stderr = self.run_command(args)
        self.assertEqual(rc, 2)
        self.assertTrue(stdout.startswith('Usage') or stdout == '')
        self.assertIn('unrecognized option', stderr)

    def run_command_missing_arg(self, args):
        rc, stdout, stderr = self.run_command(args)
        self.assertEqual(rc, 2)
        self.assertTrue(stdout.startswith('Usage') or stdout == '')
        self.assertIn('requires an argument', stderr)

    def run_command_unrecognized_tool(self, args):
        rc, stdout, stderr = self.run_command(args)
        self.assertEqual(rc, 2)
        self.assertTrue(stdout.startswith('Usage') or stdout == '')
        self.assertIn('is not a libinput command', stderr)


class TestLibinputCommand(TestLibinputTool):
    subtool = None

    def test_help(self):
        rc, stdout, stderr = self.run_command(['--help'])
        self.assertEqual(rc, 0)
        self.assertTrue(stdout.startswith('Usage:'))
        self.assertEqual(stderr, '')

    def test_version(self):
        rc, stdout, stderr = self.run_command(['--version'])
        self.assertEqual(rc, 0)
        self.assertTrue(stdout.startswith('1'))
        self.assertEqual(stderr, '')

    def test_invalid_arguments(self):
        self.run_command_unrecognized_option(['--banana'])
        self.run_command_unrecognized_option(['--foo'])
        self.run_command_unrecognized_option(['--quiet'])
        self.run_command_unrecognized_option(['--verbose'])
        self.run_command_unrecognized_option(['--quiet', 'foo'])

    def test_invalid_tools(self):
        self.run_command_unrecognized_tool(['foo'])
        self.run_command_unrecognized_tool(['debug'])
        self.run_command_unrecognized_tool(['foo', '--quiet'])


class TestToolWithOptions(object):
    options = {
        'pattern': ['sendevents'],
        # enable/disable options
        'enable-disable': [
            'tap',
            'drag',
            'drag-lock',
            'middlebutton',
            'natural-scrolling',
            'left-handed',
            'dwt'
        ],
        # options with distinct values
        'enums': {
            'set-click-method': ['none', 'clickfinger', 'buttonareas'],
            'set-scroll-method': ['none', 'twofinger', 'edge', 'button'],
            'set-profile': ['adaptive', 'flat'],
            'set-tap-map': ['lrm', 'lmr'],
        },
        # options with a range
        'ranges': {
            'set-speed': (float, -1.0, +1.0),
        }
    }

    def test_udev_seat(self):
        self.run_command_missing_arg(['--udev'])
        self.run_command_success(['--udev', 'seat0'])
        self.run_command_success(['--udev', 'seat1'])

    @unittest.skipIf(os.environ.get('UDEV_NOT_AVAILABLE'), "udev required")
    def test_device(self):
        self.run_command_missing_arg(['--device'])
        self.run_command_success(['--device', '/dev/input/event0'])
        self.run_command_success(['--device', '/dev/input/event1'])
        self.run_command_success(['/dev/input/event0'])

    def test_options_pattern(self):
        for option in self.options['pattern']:
            self.run_command_success(['--disable-{}'.format(option), '*'])
            self.run_command_success(['--disable-{}'.format(option), 'abc*'])

    def test_options_enable_disable(self):
        for option in self.options['enable-disable']:
            self.run_command_success(['--enable-{}'.format(option)])
            self.run_command_success(['--disable-{}'.format(option)])

    def test_options_enums(self):
        for option, values in self.options['enums'].items():
            for v in values:
                self.run_command_success(['--{}'.format(option), v])
                self.run_command_success(['--{}={}'.format(option, v)])

    def test_options_ranges(self):
        for option, values in self.options['ranges'].items():
            range_type, minimum, maximum = values
            self.assertEqual(range_type, float)
            step = (maximum - minimum) / 10.0
            value = minimum
            while value < maximum:
                self.run_command_success(['--{}'.format(option), str(value)])
                self.run_command_success(['--{}={}'.format(option, value)])
                value += step
            self.run_command_success(['--{}'.format(option), str(maximum)])
            self.run_command_success(['--{}={}'.format(option, maximum)])

    def test_apply_to(self):
        self.run_command_missing_arg(['--apply-to'])
        self.run_command_success(['--apply-to', '*foo*'])
        self.run_command_success(['--apply-to', 'foobar'])
        self.run_command_success(['--apply-to', 'any'])


class TestDebugEvents(TestToolWithOptions, TestLibinputTool):
    subtool = 'debug-events'

    def test_verbose_quiet(self):
        rc, stdout, stderr = self.run_command(['--verbose'])
        self.assertEqual(rc, 0)
        rc, stdout, stderr = self.run_command(['--quiet'])
        self.assertEqual(rc, 0)
        rc, stdout, stderr = self.run_command(['--verbose', '--quiet'])
        self.assertEqual(rc, 0)
        rc, stdout, stderr = self.run_command(['--quiet', '--verbose'])
        self.assertEqual(rc, 0)

    def test_invalid_arguments(self):
        self.run_command_unrecognized_option(['--banana'])
        self.run_command_unrecognized_option(['--foo'])
        self.run_command_unrecognized_option(['--version'])

    def test_multiple_devices(self):
        self.run_command_success(['--device', '/dev/input/event0', '/dev/input/event1'])
        # same event path multiple times? meh, your problem
        self.run_command_success(['--device', '/dev/input/event0', '/dev/input/event0'])
        self.run_command_success(['/dev/input/event0', '/dev/input/event1'])

    def test_too_many_devices(self):
        # Too many arguments just bails with the usage message
        rc, stdout, stderr = self.run_command(['/dev/input/event0'] * 61)
        self.assertEqual(rc, 2, msg=(stdout, stderr))


class TestDebugGUI(TestToolWithOptions, TestLibinputTool):
    subtool = 'debug-gui'

    @classmethod
    def setUpClass(cls):
        # This is set by meson
        debug_gui_enabled = @MESON_ENABLED_DEBUG_GUI@ # noqa
        if not debug_gui_enabled:
            raise unittest.SkipTest()

        if not os.getenv('DISPLAY') and not os.getenv('WAYLAND_DISPLAY'):
            raise unittest.SkipTest()

        # 77 means gtk_init() failed, which is probably because you can't
        # connect to the display server.
        rc, _, _ = run_command([TestLibinputTool.libinput_tool, cls.subtool, '--help'])
        if rc == 77:
            raise unittest.SkipTest()

    def test_verbose_quiet(self):
        rc, stdout, stderr = self.run_command(['--verbose'])
        self.assertEqual(rc, 0)

    def test_invalid_arguments(self):
        self.run_command_unrecognized_option(['--quiet'])
        self.run_command_unrecognized_option(['--banana'])
        self.run_command_unrecognized_option(['--foo'])
        self.run_command_unrecognized_option(['--version'])


class TestRecord(TestLibinputTool):
    subtool = 'record'

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.outfile = Path(self.tmpdir.name, 'record.out')

    def tearDown(self):
        self.tmpdir.cleanup()

    def test_args(self):
        self.run_command_success(['--help'])
        self.run_command_success(['--show-keycodes'])
        self.run_command_success(['--with-libinput'])

    def test_multiple_deprecated(self):
        # this arg is deprecated and a noop
        self.run_command_success(['--multiple'])

    def test_all(self):
        self.run_command_success(['--all', '-o', self.outfile])

    def test_autorestart(self):
        self.run_command_success(['--autorestart=2'])

    def test_outfile(self):
        self.run_command_success(['-o', self.outfile])
        self.run_command_success(['--output-file', self.outfile])
        self.run_command_success(['--output-file={}'.format(self.outfile)])

    def test_device_single(self):
        self.run_command_success(['/dev/input/event0'])

    def test_device_multiple(self):
        self.run_command_success(['-o', self.outfile, '/dev/input/event0', '/dev/input/event1'])


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Verify a libinput tool\'s option parsing')
    parser.add_argument('--tool-path', metavar='/path/to/builddir/libinput',
                        type=str,
                        help='Path to the libinput tool in the builddir')
    parser.add_argument('--verbose', action='store_true')
    args, remainder = parser.parse_known_args()
    if args.tool_path is not None:
        TestLibinputTool.libinput_tool = args.tool_path
    verbosity = 1
    if args.verbose:
        verbosity = 3

    argv = [sys.argv[0], *remainder]
    unittest.main(verbosity=verbosity, argv=argv)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#

import os
import sys
import argparse
import subprocess
try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print('Error: {}'.format(str(e)), file=sys.stderr)
    print('One or more python modules are missing. Please install those '
          'modules and re-run this tool.')
    sys.exit(1)


DEFAULT_HWDB_FILE = '/usr/lib/udev/hwdb.d/60-evdev.hwdb'
OVERRIDE_HWDB_FILE = '/etc/udev/hwdb.d/99-touchpad-fuzz-override.hwdb'


class tcolors:
    GREEN = '\033[92m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    NORMAL = '\033[0m'


def print_bold(msg, **kwargs):
    print(tcolors.BOLD + msg + tcolors.NORMAL, **kwargs)


def print_green(msg, **kwargs):
    print(tcolors.BOLD + tcolors.GREEN + msg + tcolors.NORMAL, **kwargs)


def print_red(msg, **kwargs):
    print(tcolors.BOLD + tcolors.RED + msg + tcolors.NORMAL, **kwargs)


class InvalidConfigurationError(Exception):
    pass


class InvalidDeviceError(Exception):
    pass


class Device(libevdev.Device):
    def __init__(self, path):
        if path is None:
            self.path = self.find_touch_device()
        else:
            self.path = path

        fd = open(self.path, 'rb')
        super().__init__(fd)
        context = pyudev.Context()
        self.udev_device = pyudev.Devices.from_device_file(context, self.path)

    def find_touch_device(self):
        context = pyudev.Context()
        for device in context.list_devices(subsystem='input'):
            if not device.get('ID_INPUT_TOUCHPAD', 0):
                continue

            if not device.device_node or \
                    not device.device_node.startswith('/dev/input/event'):
                continue

            return device.device_node

        print('Unable to find a touch device.', file=sys.stderr)
        sys.exit(1)

    def check_property(self):
        '''Return a tuple of (xfuzz, yfuzz) with the fuzz as set in the libinput
           property. Returns None if the property doesn't exist'''

        axes = {
            0x00: self.udev_device.get('LIBINPUT_FUZZ_00'),
            0x01: self.udev_device.get('LIBINPUT_FUZZ_01'),
            0x35: self.udev_device.get('LIBINPUT_FUZZ_35'),
            0x36: self.udev_device.get('LIBINPUT_FUZZ_36'),
        }

        if axes[0x35] is not None:
            if axes[0x35] != axes[0x00]:
                print_bold('WARNING: fuzz mismatch ABS_X: {}, ABS_MT_POSITION_X: {}'.format(axes[0x00], axes[0x35]))

        if axes[0x36] is not None:
            if axes[0x36] != axes[0x01]:
                print_bold('WARNING: fuzz mismatch ABS_Y: {}, ABS_MT_POSITION_Y: {}'.format(axes[0x01], axes[0x36]))

        xfuzz = axes[0x35] or axes[0x00]
        yfuzz = axes[0x36] or axes[0x01]

        if xfuzz is None and yfuzz is None:
            return None

        if ((xfuzz is not None and yfuzz is None) or
                (xfuzz is None and yfuzz is not None)):
            raise InvalidConfigurationError('fuzz should be set for both axes')

        return (xfuzz, yfuzz)

    def check_axes(self):
        '''
        Returns a tuple of (xfuzz, yfuzz) with the fuzz as set on the device
        axis. Returns None if no fuzz is set.
        '''
        if not self.has(libevdev.EV_ABS.ABS_X) or not self.has(libevdev.EV_ABS.ABS_Y):
            raise InvalidDeviceError('device does not have x/y axes')

        if self.has(libevdev.EV_ABS.ABS_MT_POSITION_X) != self.has(libevdev.EV_ABS.ABS_MT_POSITION_Y):
            raise InvalidDeviceError('device does not have both multitouch axes')

        xfuzz = (self.absinfo[libevdev.EV_ABS.ABS_X].fuzz or
                 self.absinfo[libevdev.EV_ABS.ABS_MT_POSITION_X].fuzz)
        yfuzz = (self.absinfo[libevdev.EV_ABS.ABS_Y].fuzz or
                 self.absinfo[libevdev.EV_ABS.ABS_MT_POSITION_Y].fuzz)

        if xfuzz == 0 and yfuzz == 0:
            return None

        return (xfuzz, yfuzz)


def print_fuzz(what, fuzz):
    print('  Checking {}... '.format(what), end='')
    if fuzz is None:
        print('not set')
    elif fuzz == (0, 0):
        print('is zero')
    else:
        print('x={} y={}'.format(*fuzz))


def handle_existing_entry(device, fuzz):
    # This is getting messy because we don't really know where the entry
    # could be or how the match rule looks like. So we just check the
    # default location only.
    # For the match comparison, we search for the property value in the
    # file. If there is more than one entry that uses the same
    # overrides this will generate false positives.
    # If the lines aren't in the same order in the file, it'll be a false
    # negative.
    overrides = {
        0x00: device.udev_device.get('EVDEV_ABS_00'),
        0x01: device.udev_device.get('EVDEV_ABS_01'),
        0x35: device.udev_device.get('EVDEV_ABS_35'),
        0x36: device.udev_device.get('EVDEV_ABS_36'),
    }

    has_existing_rules = False
    for key, value in overrides.items():
        if value is not None:
            has_existing_rules = True
            break
    if not has_existing_rules:
        return False

    print_red('Error! ', end='')
    print('This device already has axis overrides defined')
    print('')
    print_bold('Searching for existing override...')

    # Construct a template that looks like a hwdb entry (values only) from
    # the udev property values
    template = [' EVDEV_ABS_00={}'.format(overrides[0x00]),
                ' EVDEV_ABS_01={}'.format(overrides[0x01])]
    if overrides[0x35] is not None:
        template += [' EVDEV_ABS_35={}'.format(overrides[0x35]),
                     ' EVDEV_ABS_36={}'.format(overrides[0x36])]

    print('Checking in {}... '.format(OVERRIDE_HWDB_FILE), end='')
    entry, prefix, lineno = check_file_for_lines(OVERRIDE_HWDB_FILE, template)
    if entry is not None:
        print_green('found')
        print('The existing hwdb entry can be overwritten')
        return False
    else:
        print_red('not found')
        print('Checking in {}... '.format(DEFAULT_HWDB_FILE, template), end='')
        entry, prefix, lineno = check_file_for_lines(DEFAULT_HWDB_FILE, template)
        if entry is not None:
            print_green('found')
        else:
            print_red('not found')
            print('The device has a hwdb override defined but it\'s not where I expected it to be.')
            print('Please look at the libinput documentation for more details.')
            print('Exiting now.')
            return True

    print_bold('Probable entry for this device found in line {}:'.format(lineno))
    print('\n'.join(prefix + entry))
    print('')

    print_bold('Suggested new entry for this device:')
    new_entry = []
    for i in range(0, len(template)):
        parts = entry[i].split(':')
        while len(parts) < 4:
            parts.append('')
        parts[3] = str(fuzz)
        new_entry.append(':'.join(parts))
    print('\n'.join(prefix + new_entry))
    print('')

    # Not going to overwrite the 60-evdev.hwdb entry with this program, too
    # risky. And it may not be our device match anyway.
    print_bold('You must now:')
    print('\n'.join((
        '1. Check the above suggestion for sanity. Does it match your device?',
        '2. Open {} and amend the existing entry'.format(DEFAULT_HWDB_FILE),
        '   as recommended above',
        '',
        '   The property format is:',
        '    EVDEV_ABS_00=min:max:resolution:fuzz',
        '',
        '   Leave the entry as-is and only add or amend the fuzz value.',
        '   A non-existent value can be skipped, e.g. this entry sets the ',
        '   resolution to 32 and the fuzz to 8',
        '    EVDEV_ABS_00=::32:8',
        '',
        '3. Save the edited file',
        '4. Say Y to the next prompt')))

    cont = input('Continue? [Y/n] ')
    if cont == 'n':
        raise KeyboardInterrupt

    if test_hwdb_entry(device, fuzz):
        print_bold('Please test the new fuzz setting by restarting libinput')
        print_bold('Then submit a pull request for this hwdb entry change to '
                   'to systemd at http://github.com/systemd/systemd')
    else:
        print_bold('The new fuzz setting did not take effect.')
        print_bold('Did you edit the correct file?')
        print('Please look at the libinput documentation for more details.')
        print('Exiting now.')

    return True


def reload_and_trigger_udev(device):
    import time

    print('Running udevadm hwdb --update')
    subprocess.run(['udevadm', 'hwdb', '--update'], check=True)
    syspath = device.path.replace('/dev/input/', '/sys/class/input/')
    time.sleep(1)
    print('Running udevadm trigger {}'.format(syspath))
    subprocess.run(['udevadm', 'trigger', syspath], check=True)
    time.sleep(1)


def test_hwdb_entry(device, fuzz):
    reload_and_trigger_udev(device)
    print_bold('Testing... ', end='')

    d = Device(device.path)
    f = d.check_axes()
    if fuzz == f[0] and fuzz == f[1]:
        print_green('Success')
        return True
    else:
        print_red('Error')
        return False


def check_file_for_lines(path, template):
    '''
    Checks file at path for the lines given in template. If found, the
    return value is a tuple of the matching lines and the prefix (i.e. the
    two lines before the matching lines)
    '''
    try:
        lines = [l[:-1] for l in open(path).readlines()]
        idx = -1
        try:
            while idx < len(lines) - 1:
                idx += 1
                line = lines[idx]
                if not line.startswith(' EVDEV_ABS_00'):
                    continue
                if lines[idx:idx + len(template)] != template:
                    continue

                return (lines[idx:idx + len(template)], lines[idx - 2:idx], idx)

        except IndexError:
            pass
    except FileNotFoundError:
        pass

    return (None, None, None)


def write_udev_rule(device, fuzz):
    '''Write out a udev rule that may match the device, run udevadm trigger and
    check if the udev rule worked. Of course, there's plenty to go wrong...
    '''
    print('')
    print_bold('Guessing a udev rule to overwrite the fuzz')

    # Some devices match better on pvr, others on pn, so we get to try both. yay
    modalias = open('/sys/class/dmi/id/modalias').readlines()[0]
    ms = modalias.split(':')
    svn, pn, pvr = None, None, None
    for m in ms:
        if m.startswith('svn'):
            svn = m
        elif m.startswith('pn'):
            pn = m
        elif m.startswith('pvr'):
            pvr = m

    # Let's print out both to inform and/or confuse the user
    template = '\n'.join(('# {} {}',
                          'evdev:name:{}:dmi:*:{}*:{}*:',
                          ' EVDEV_ABS_00=:::{}',
                          ' EVDEV_ABS_01=:::{}',
                          ' EVDEV_ABS_35=:::{}',
                          ' EVDEV_ABS_36=:::{}',
                          ''))
    rule1 = template.format(svn[3:], device.name, device.name, svn, pvr, fuzz, fuzz, fuzz, fuzz)
    rule2 = template.format(svn[3:], device.name, device.name, svn, pn, fuzz, fuzz, fuzz, fuzz)

    print('Full modalias is: {}'.format(modalias))
    print()
    print_bold('Suggested udev rule, option 1:')
    print(rule1)
    print()
    print_bold('Suggested udev rule, option 2:')
    print(rule2)
    print('')

    # The weird hwdb matching behavior means we match on the least specific
    # rule (i.e. most wildcards) first although that was supposed to be fixed in
    # systemd 3a04b789c6f1.
    # Our rule uses dmi strings and will be more specific than what 60-evdev.hwdb
    # already has. So we basically throw up our hands because we can't do anything
    # then.
    if handle_existing_entry(device, fuzz):
        return

    while True:
        print_bold('Wich rule do you want to to test? 1 or 2? ', end='')
        yesno = input('Ctrl+C to exit ')

        if yesno == '1':
            rule = rule1
            break
        elif yesno == '2':
            rule = rule2
            break

    fname = OVERRIDE_HWDB_FILE
    try:
        fd = open(fname, 'x')
    except FileExistsError:
        yesno = input('File {} exists, overwrite? [Y/n] '.format(fname))
        if yesno.lower == 'n':
            return

        fd = open(fname, 'w')

    fd.write('# File generated by libinput measure fuzz\n\n')
    fd.write(rule)
    fd.close()

    if test_hwdb_entry(device, fuzz):
        print('Your hwdb override file is in {}'.format(fname))
        print_bold('Please test the new fuzz setting by restarting libinput')
        print_bold('Then submit a pull request for this hwdb entry to '
                   'systemd at http://github.com/systemd/systemd')
    else:
        print('The hwdb entry failed to apply to the device.')
        print('Removing hwdb file again.')
        os.remove(fname)
        reload_and_trigger_udev(device)
        print_bold('What now?')
        print('1. Re-run this program and try the other suggested udev rule. If that fails,')
        print('2. File a bug with the suggested udev rule at http://github.com/systemd/systemd')


def main(args):
    parser = argparse.ArgumentParser(
        description='Print fuzz settings and/or suggest udev rules for the fuzz to be adjusted.'
    )
    parser.add_argument('path', metavar='/dev/input/event0',
                        nargs='?', type=str, help='Path to device (optional)')
    parser.add_argument('--fuzz', type=int, help='Suggested fuzz')
    args = parser.parse_args()

    try:
        device = Device(args.path)
        print_bold('Using {}: {}'.format(device.name, device.path))

        fuzz = device.check_property()
        print_fuzz('udev property', fuzz)

        fuzz = device.check_axes()
        print_fuzz('axes', fuzz)

        userfuzz = args.fuzz
        if userfuzz is not None:
            write_udev_rule(device, userfuzz)

    except PermissionError:
        print('Permission denied, please re-run as root')
    except InvalidConfigurationError as e:
        print('Error: {}'.format(e))
    except InvalidDeviceError as e:
        print('Error: {}'.format(e))
    except KeyboardInterrupt:
        print('Exited on user request')


if __name__ == '__main__':
    main(sys.argv)
